import Control.Parallel.Strategies (parList, using, rseq)
import Data.Binary (Word32)
import Data.Bits (Bits (shiftL, shiftR, (.&.)))
import Data.List (permutations)
import Data.List.Split (chunksOf)
import System.Environment (getArgs, getProgName)
import System.IO (hPutStrLn, stderr)
import System.Exit (exitFailure)
import GHC.Conc (numCapabilities)

type ShiftFn = Word32 -> Int -> Word32

-- generates an nxn board of queens arranged in a diagonal
-- from the top-right to the bottom-left
-- for example, for n = 4 the following board:
--    0 0 0 1
--    0 0 1 0
--    0 1 0 0
--    1 0 0 0
-- the board is represented as a list of numbers
-- So the above example is represented as:
-- [0b0001, 0b0010, 0b0100, 0b1000] (== [1, 2, 4, 8])
diagonalOfQueens :: Int -> [Word32]
diagonalOfQueens n = [shiftL 1 i | i <- [0 .. (n -1)]]

-- given a row and list of subsequent rows,
-- checks if the first row has any diagonal conflicts with the subsequent rows
-- Each row down the first row shifts to the left or right by 1. Set by `shiftFn`
hasConflictWithFirstRow :: ShiftFn -> Word32 -> [Word32] -> Int -> Bool
hasConflictWithFirstRow _ _ [] _ = False
hasConflictWithFirstRow shiftFn queen (x : xs) offset =
  ((queen `shiftFn` offset) .&. x) /= 0 || hasConflictWithFirstRow shiftFn queen xs (offset + 1)

-- Checks if a board configuration has any diagonal conflicts
-- Checks if each row has any conflicts with the rows below it
hasAnyDiagonalConflict :: [Word32] -> Bool
hasAnyDiagonalConflict board =
  isDiagonalConflict shiftL board || isDiagonalConflict shiftR board
  where
    isDiagonalConflict _ [] = False
    isDiagonalConflict dir (x : xs) =
      hasConflictWithFirstRow dir x xs 1 || isDiagonalConflict dir xs

-- Haskell's laziness seems to give us the same performance if we use
-- filter and then `length` vs mapping to 1s and 0s and then summing
countValidBoards :: [[Word32]] -> Int
countValidBoards = length . filter (not . hasAnyDiagonalConflict)

countSolutions :: [Word32] -> Int -> Int
countSolutions board n =
  let
    unmovingRow = board !! n
    otherRows = take n board ++ drop (n + 1) board
    boards = map (unmovingRow :) (permutations otherRows)
  in
    countValidBoards boards

-- gets the number of solutions for a given n
nQueensSolutionCount :: Int -> Int
nQueensSolutionCount n =
  let diagonal = diagonalOfQueens n
      solutions = map (countSolutions diagonal) [0..(n-1)] `using` parList rseq
    in sum solutions

main :: IO ()
main = do
  args <- getArgs
  case args of
    [n] ->
      putStrLn $ "There are " ++ show (nQueensSolutionCount (read n)) ++ " solutions for " ++ show n ++ " queens."
    _ -> do
      pn <- getProgName
      hPutStrLn stderr ("Usage: " ++ pn ++ " <Int>")
      exitFailure
